from __future__ import annotations

from collections.abc import Callable, Coroutine
from concurrent.futures import Future
from functools import wraps
from inspect import iscoroutinefunction
from logging import getLogger
from threading import Lock, current_thread
from typing import Any, Literal, Protocol, cast, overload

from PySide6.QtCore import QObject, QRunnable, QThreadPool, Signal, Slot
from PySide6.QtWidgets import QApplication
from vsengine.loops import EventLoop, get_loop

type _CoroutineFunc[**P, R] = Callable[P, Coroutine[Any, Any, R]]
type _Func[**P, R] = Callable[P, R]

_logger = getLogger(__name__)


class QtEventLoop(QObject, EventLoop):
    """Qt event loop adapter for vsview."""

    _invoke = Signal(int)

    def attach(self) -> None:
        self._lock = Lock()
        self._counter = 0
        self._pending = dict[int, Callable[[], None]]()
        self._invoke.connect(self._on_invoke)

    def detach(self) -> None:
        self.wait_for_threads(100)
        self._invoke.disconnect(self._on_invoke)
        self._pending.clear()

    @Slot(int)
    def _on_invoke(self, task_id: int) -> None:
        with self._lock:
            wrapper = self._pending.pop(task_id, None)
        if wrapper:
            wrapper()

    def from_thread[**P, R](self, func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> Future[R]:
        """Schedule func to run on the main Qt thread."""
        fut = Future[R]()

        def wrapper() -> None:
            if not fut.set_running_or_notify_cancel():
                return
            try:
                result = func(*args, **kwargs)
            except BaseException as e:
                _logger.debug(e, exc_info=True)
                fut.set_exception(e)
            else:
                fut.set_result(result)

        with self._lock:
            task_id = self._counter
            self._counter += 1
            self._pending[task_id] = wrapper

        self._invoke.emit(task_id)

        return fut

    def to_thread[**P, R](self, func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> Future[R]:
        """Run func in Qt's global thread pool."""
        fut = Future[R]()

        def wrapper() -> None:
            if not fut.set_running_or_notify_cancel():
                return
            try:
                result = func(*args, **kwargs)
            except BaseException as e:
                _logger.error(e)
                _logger.debug(e, exc_info=True)
                fut.set_exception(e)
            else:
                fut.set_result(result)

        QThreadPool.globalInstance().start(QRunnable.create(wrapper))
        return fut

    def to_thread_named[**P, R](self, name: str, func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> Future[R]:
        """Run func in Qt's global thread pool with a custom thread name."""
        fut = Future[R]()

        def wrapper() -> None:
            current_thread().name = name
            if not fut.set_running_or_notify_cancel():
                return
            try:
                result = func(*args, **kwargs)
            except BaseException as e:
                _logger.error(e)
                _logger.debug(e, exc_info=True)
                fut.set_exception(e)
            else:
                fut.set_result(result)

        QThreadPool.globalInstance().start(QRunnable.create(wrapper))
        return fut

    def wait_for_threads(self, timeout_ms: int = 500) -> None:
        """Wait for all background threads in the global thread pool to finish."""

        pool = QThreadPool.globalInstance()

        for _ in range(max(1, timeout_ms // 10)):
            if pool.activeThreadCount() == 0:
                break
            pool.waitForDone(10)
            QApplication.processEvents()

        # Final flush to process any signals from threads that just finished
        QApplication.processEvents()


class _DecoratorFuture(Protocol):
    """Protocol for decorators that wrap a function and return a Future."""

    @overload
    def __call__[**P, R](self, func: _CoroutineFunc[P, R]) -> Callable[P, Future[R]]: ...
    @overload
    def __call__[**P, R](self, func: _Func[P, R]) -> Callable[P, Future[R]]: ...


class _DecoratorDirect(Protocol):
    """Protocol for decorators that wrap a function and return the result directly."""

    @overload
    def __call__[**P, R](self, func: _CoroutineFunc[P, R]) -> Callable[P, R]: ...
    @overload
    def __call__[**P, R](self, func: _Func[P, R]) -> Callable[P, R]: ...


@overload
def run_in_loop[**P, R](func: _CoroutineFunc[P, R]) -> Callable[P, Future[R]]: ...
@overload
def run_in_loop[**P, R](func: _Func[P, R]) -> Callable[P, Future[R]]: ...
@overload
def run_in_loop(*, return_future: Literal[True]) -> _DecoratorFuture: ...
@overload
def run_in_loop(*, return_future: Literal[False]) -> _DecoratorDirect: ...
@overload
def run_in_loop(*, return_future: bool) -> _DecoratorFuture | _DecoratorDirect: ...
def run_in_loop(func: Any = None, *, return_future: bool = True) -> Any:
    """
    Decorator. Executes the decorated function within the `QtEventLoop` (Main Thread).

    Args:
        func: The function to wrap (when used as `@run_in_loop` without parens)
        return_future: If False, blocks and returns R directly.

    Returns:
        A future object or the result directly, depending on return_future.

    Usage:
    ```python
    @run_in_loop
    def my_func(): ...


    @run_in_loop(return_future=False)
    def my_blocking_func(): ...
    ```
    """

    def decorator(fn: Any) -> Any:
        @wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            loop = cast(QtEventLoop, get_loop())

            if iscoroutinefunction(fn):

                def run_coro() -> Any:
                    import asyncio

                    coro = fn(*args, **kwargs)
                    try:
                        return asyncio.run(coro)
                    except RuntimeError:
                        return asyncio.run_coroutine_threadsafe(coro, asyncio.get_running_loop()).result()

                fut = loop.from_thread(run_coro)
            else:
                # Delegate to from_thread to marshal execution to the main loop
                fut = loop.from_thread(fn, *args, **kwargs)

            return fut if return_future else fut.result()

        return wrapper

    return decorator if func is None else decorator(func)


@overload
def run_in_background[**P, R](func: _CoroutineFunc[P, R]) -> Callable[P, Future[R]]: ...
@overload
def run_in_background[**P, R](func: _Func[P, R]) -> Callable[P, Future[R]]: ...
@overload
def run_in_background(*, name: str) -> _DecoratorFuture: ...
def run_in_background(func: Any = None, *, name: str | None = None) -> Any:
    """
    Executes the decorated function in a background thread (via QThreadPool)
    using the `QtEventLoop`'s `to_thread` logic.

    Args:
        func: The function to wrap (when used as `@run_in_background` without parens)
        name: Optional thread name for logging (when used as `@run_in_background(name="...")`)

    Returns:
        A future object representing the result of the execution.

    Usage:
    ```python
    @run_in_background
    def my_func(): ...


    @run_in_background(name="MyWorker")
    def my_named_func(): ...
    ```
    """

    def decorator(fn: Any) -> Any:
        @wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            loop = cast(QtEventLoop, get_loop())

            if iscoroutinefunction(fn):

                def run_coro() -> Any:
                    import asyncio

                    coro = fn(*args, **kwargs)
                    try:
                        return asyncio.run(coro)
                    except RuntimeError:
                        return asyncio.run_coroutine_threadsafe(coro, asyncio.get_running_loop()).result()

                return loop.to_thread(run_coro) if name is None else loop.to_thread_named(name, run_coro)

            return (
                loop.to_thread(fn, *args, **kwargs) if name is None else loop.to_thread_named(name, fn, *args, **kwargs)
            )

        return wrapper

    return decorator if func is None else decorator(func)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import timedelta
from typing import Annotated, Any, final
from uuid import UUID, uuid4

from jetpytools import fallback
from pydantic import BaseModel, BeforeValidator, ConfigDict, Field, field_serializer, model_validator
from PySide6.QtGui import QColor

from vsview.api import Time, VideoOutputProxy


class UUIDModel(BaseModel):
    id: UUID = Field(default_factory=uuid4, repr=False, init=False)

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, UUIDModel) else NotImplemented


class AbstractRange[T](ABC, UUIDModel):
    """
    Base class for a scening range, which can be defined in frames or timestamps.

    A range consists of a start point, an optional end point (defaults to start), and an optional label.
    """

    start: T
    """The start point of the range."""
    end: T | None = None
    """The end point of the range. If None, the range represents a single point (start)."""
    label: str = ""
    """An optional description for this range."""

    @abstractmethod
    def as_frames(self, v: VideoOutputProxy) -> tuple[int, int]:
        """
        Convert the range to a tuple of (start_frame, end_frame).

        Args:
            v: The video output proxy used for conversion.

        Returns:
            tuple[int, int]: The start and end frames.
        """
        ...

    @abstractmethod
    def as_times(self, v: VideoOutputProxy) -> tuple[Time, Time]:
        """
        Convert the range to a tuple of (Time, Time).

        Args:
            v: The video output proxy used for conversion.

        Returns:
            tuple[Time, Time]: The start and end times.
        """
        ...

    @abstractmethod
    def from_frames(self, s: int | None, e: int | None, v: VideoOutputProxy) -> None:
        """
        Update the range boundaries from frame numbers.

        Args:
            s: The new start frame.
            e: The new end frame.
            v: The video output proxy used for conversion.
        """
        ...

    @abstractmethod
    def from_times(self, s: timedelta | None, e: timedelta | None, v: VideoOutputProxy) -> None:
        """
        Update the range boundaries from timedelta objects.

        Args:
            s: The new start time.
            e: The new end time.
            v: The video output proxy used for conversion.
        """
        ...

    def to_tuple(self) -> tuple[T, T]:
        """
        Return a tuple of (start, end), where end defaults to start if None.
        """
        return self.start, fallback(self.end, self.start)


class RangeFrame(AbstractRange[int], UUIDModel):
    """A range defined by frame numbers."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="before")
    @classmethod
    def _transform_input(cls, data: Any) -> Any:
        if isinstance(data, int):
            return {"start": data}
        if isinstance(data, (list, tuple)) and len(data) >= 2:
            return {"start": data[0], "end": data[1]}
        return data

    def as_frames(self, v: VideoOutputProxy) -> tuple[int, int]:
        return self.start, (self.end if self.end is not None else self.start)

    def as_times(self, v: VideoOutputProxy) -> tuple[Time, Time]:
        s, e = self.as_frames(v)
        return v.frame_to_time(s), v.frame_to_time(e)

    def from_frames(self, s: int | None, e: int | None, v: VideoOutputProxy) -> None:
        if s is not None:
            self.start = s
        if e is not None:
            self.end = e

    def from_times(self, s: timedelta | None, e: timedelta | None, v: VideoOutputProxy) -> None:
        if s is not None:
            self.start = v.time_to_frame(s)
        if e is not None:
            self.end = v.time_to_frame(e)


class RangeTime(AbstractRange[timedelta], UUIDModel):
    """A range defined by timestamps (timedelta)."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="before")
    @classmethod
    def _transform_input(cls, data: Any) -> Any:
        def _to_td(ts: Any) -> timedelta:
            if isinstance(ts, timedelta):
                return ts
            h, m, s = str(ts).split(":")
            return timedelta(hours=int(h), minutes=int(m), seconds=float(s))

        if isinstance(data, (str, timedelta)):
            return {"start": _to_td(data)}
        if isinstance(data, (list, tuple)) and len(data) >= 2:
            return {"start": _to_td(data[0]), "end": _to_td(data[1])}
        return data

    def as_frames(self, v: VideoOutputProxy) -> tuple[int, int]:
        s, e = self.as_times(v)
        return v.time_to_frame(s), v.time_to_frame(e)

    def as_times(self, v: VideoOutputProxy) -> tuple[Time, Time]:
        s = self.start
        e = self.end if self.end is not None else s

        return Time(seconds=s.total_seconds()), Time(seconds=e.total_seconds())

    def from_frames(self, s: int | None, e: int | None, v: VideoOutputProxy) -> None:
        if s is not None:
            self.start = v.frame_to_time(s)
        if e is not None:
            self.end = v.frame_to_time(e)

    def from_times(self, s: timedelta | None, e: timedelta | None, v: VideoOutputProxy) -> None:
        if s is not None:
            self.start = s
        if e is not None:
            self.end = e


@final
class UnifiedRange:
    """A helper wrapper providing a unified interface for frame and time-based ranges."""

    def __init__(
        self,
        r: RangeFrame | RangeTime,
        frame_to_time: Callable[[int], Time],
        time_to_frame: Callable[[timedelta], int],
    ) -> None:
        self._r = r
        self._frame_to_time = frame_to_time
        self._time_to_frame = time_to_frame
        self.label = self._r.label

    def as_frames(self) -> tuple[int, int]:
        """Convert the range to a tuple of (start_frame, end_frame)."""
        if isinstance(self._r, RangeFrame):
            return self._r.to_tuple()

        s, e = self._r.to_tuple()

        return self._time_to_frame(s), self._time_to_frame(e)

    def as_times(self) -> tuple[Time, Time]:
        """Convert the range to a tuple of (Time, Time)."""
        if isinstance(self._r, RangeTime):
            s, e = self._r.to_tuple()
            return Time(seconds=s.total_seconds()), Time(seconds=e.total_seconds())

        s, e = self._r.to_tuple()

        return self._frame_to_time(s), self._frame_to_time(e)


class SceneRow(UUIDModel):
    """Represents a collection of ranges grouped under a single name and color."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    color: Annotated[QColor, BeforeValidator(lambda v: v if isinstance(v, QColor) else QColor(v))]
    name: str
    checked_outputs: set[int] = Field(default_factory=set)
    display: bool = True

    ranges: list[RangeFrame | RangeTime] | list[RangeFrame] | list[RangeTime] = Field(default_factory=list)

    @property
    def notch_id(self) -> str:
        from .plugin import PLUGIN_IDENTIFIER

        return ".".join([PLUGIN_IDENTIFIER, str(self.id)])

    @field_serializer("color")
    def serialize_color(self, color: QColor) -> str:
        return color.name()



from logging import getLogger

logger = getLogger(__name__)


def gc_collect() -> None:
    logger.debug("Running garbage collection")
    import gc

    for i in range(3):
        gc.collect(generation=i)

    for _ in range(3):
        gc.collect()



import sys

from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import QApplication, QHeaderView, QStyle, QTableWidget, QTableWidgetItem, QVBoxLayout, QWidget


class StandardIconBrowser(QWidget):
    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle("PySide6 Standard Pixmaps")
        self.resize(450, 600)

        layout = QVBoxLayout(self)

        self.table = QTableWidget()
        self.table.setColumnCount(2)
        self.table.setHorizontalHeaderLabels(["Icon", "Enum Name"])

        self.table.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        self.table.verticalHeader().setVisible(False)
        self.table.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)

        self.table.setIconSize(QSize(32, 32))

        layout.addWidget(self.table)

        self.load_icons()

    def load_icons(self) -> None:
        style = self.style()

        for member in QStyle.StandardPixmap:
            if member.name == "SP_CustomBase":
                continue

            row_idx = self.table.rowCount()
            self.table.insertRow(row_idx)

            icon = style.standardIcon(member)

            icon_item = QTableWidgetItem(icon, "")
            name_item = QTableWidgetItem(member.name)
            name_item.setTextAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignLeft)

            self.table.setItem(row_idx, 0, icon_item)
            self.table.setItem(row_idx, 1, name_item)

            self.table.setRowHeight(row_idx, 40)


if __name__ == "__main__":
    app = QApplication(sys.argv)

    window = StandardIconBrowser()
    window.show()

    sys.exit(app.exec())



from __future__ import annotations

from collections.abc import Callable
from itertools import zip_longest
from logging import getLogger
from pathlib import Path
from types import get_original_bases
from typing import TYPE_CHECKING, Any, get_args, get_origin
from weakref import WeakKeyDictionary

import vapoursynth as vs
from pydantic import BaseModel
from PySide6.QtCore import QMetaObject, QObject, QRect, Signal
from PySide6.QtGui import QContextMenuEvent, QKeyEvent, QMouseEvent
from PySide6.QtWidgets import QDockWidget, QSplitter, QTabWidget, QWidget

from vsview.app.outputs import VideoOutput
from vsview.app.settings import SecretsManager, SettingsManager
from vsview.app.utils import ObjectType
from vsview.app.views.timeline import Timeline
from vsview.app.views.video import GraphicsView
from vsview.vsenv.loop import run_in_loop

if TYPE_CHECKING:
    from vsview.app.workspace.loader import LoaderWorkspace
    from vsview.app.workspace.playback import PlaybackManager

    from .api import PluginGraphicsView, VideoOutputProxy, WidgetPluginBase, _PluginBase

logger = getLogger(__name__)


class _SettingsProxy[T: BaseModel]:
    """
    Proxy that intercepts `__setattr__` on a pydantic BaseModel and auto-persists the change via the supplied callback.
    Reads are delegated transparently to the underlying model.
    """

    __slots__ = ("_model", "_on_update")

    def __init__(self, model: T, on_update: Callable[[str, Any], None]) -> None:
        object.__setattr__(self, "_model", model)
        object.__setattr__(self, "_on_update", on_update)

    def __getattr__(self, name: str) -> Any:
        attr = getattr(self._model, name)
        return _SettingsProxy(attr, lambda _, __: self._on_update(name, attr)) if isinstance(attr, BaseModel) else attr

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.__slots__:
            object.__setattr__(self, name, value)
        elif hasattr(self._model, name):
            if isinstance(value, _SettingsProxy):
                value = value._model

            setattr(self._model, name, value)
            self._on_update(name, value)
        else:
            raise AttributeError(f"{type(self._model).__name__!r} object has no attribute {name!r}")

    def __repr__(self) -> str:
        return repr(self._model)

    def __eq__(self, other: object) -> bool:
        return self._model == other._model if isinstance(other, _SettingsProxy) else self._model == other


class _PluginSettingsStore:
    def __init__(self, workspace: LoaderWorkspace[Any]) -> None:
        self._workspace = workspace
        self._caches: dict[str, WeakKeyDictionary[_PluginBase[Any, Any], BaseModel]] = {
            "global": WeakKeyDictionary(),
            "local": WeakKeyDictionary(),
        }

    @property
    def file_path(self) -> Path | None:
        from vsview.app.workspace.file import GenericFileWorkspace

        return self._workspace.content if isinstance(self._workspace, GenericFileWorkspace) else None

    def get(self, plugin: _PluginBase[Any, Any], scope: str) -> BaseModel | None:
        cache = self._caches[scope]

        if plugin in cache:
            return cache[plugin]

        model_cls: type[BaseModel] | None = getattr(plugin, f"{scope}_settings_model")
        if model_cls is None:
            return None

        # Fetch and validate
        settings = model_cls.model_validate(self._get_raw_settings(plugin.identifier, scope))

        # Resolve local settings with global fallbacks
        if scope == "local":
            from .api import LocalSettingsModel

            if isinstance(settings, LocalSettingsModel) and (global_settings := self.get(plugin, "global")):
                settings = settings.resolve(global_settings)

        cache[plugin] = settings
        return settings

    def update(self, plugin: _PluginBase[Any, Any], scope: str, **updates: Any) -> None:
        model_cls: type[BaseModel] | None = getattr(plugin, f"{scope}_settings_model")

        if model_cls is None:
            logger.warning("No model for plugin %r scope %r", plugin.identifier, scope)
            return

        # Always update against the raw (unresolved) state
        raw_data = self._get_raw_settings(plugin.identifier, scope)
        settings = model_cls.model_validate(raw_data)

        for key, value in updates.items():
            setattr(settings, key, value)

        self._set_raw_settings(plugin.identifier, scope, settings)
        self._caches[scope].pop(plugin, None)

    def invalidate(self, scope: str) -> None:
        self._caches[scope].clear()

    def _get_raw_settings(self, plugin_id: str, scope: str) -> dict[str, Any]:
        if scope == "global":
            container = SettingsManager.global_settings
        elif self.file_path is not None:
            container = SettingsManager.get_local_settings(self.file_path)
        else:
            return {}

        raw = container.plugins.get(plugin_id, {})
        return raw if isinstance(raw, dict) else raw.model_dump()

    def _set_raw_settings(self, plugin_id: str, scope: str, settings: BaseModel) -> None:
        if scope == "global":
            SettingsManager.global_settings.plugins[plugin_id] = settings
        elif self.file_path is not None:
            SettingsManager.get_local_settings(self.file_path).plugins[plugin_id] = settings


def _make_voutput_proxy(voutput: VideoOutput) -> VideoOutputProxy:
    from .api import VideoOutputProxy

    return VideoOutputProxy(
        voutput.vs_index,
        voutput.vs_name,
        voutput.vs_output,
        voutput.props,
        tuple(voutput.framedurs) if voutput.framedurs is not None else None,
        tuple(voutput.cum_durations) if voutput.cum_durations is not None else None,
        voutput.info,
    )


class _PluginAPI(QObject):
    statusMessage = Signal(str)
    globalSettingsChanged = Signal()
    localSettingsChanged = Signal(str)
    aboutToSaveGlobal = Signal()
    aboutToSaveLocal = Signal(str)

    def __init__(self, workspace: LoaderWorkspace[Any]) -> None:
        super().__init__()
        self.__workspace = workspace
        self.__settings_store: _PluginSettingsStore | None = None

        SettingsManager.signals.globalChanged.connect(self._on_global_settings_changed)
        SettingsManager.signals.localChanged.connect(self._on_local_settings_changed)
        SettingsManager.signals.aboutToSaveGlobal.connect(self.aboutToSaveGlobal.emit)
        SettingsManager.signals.aboutToSaveLocal.connect(self.aboutToSaveLocal.emit)

    @property
    def voutputs(self) -> list[VideoOutputProxy]:
        """Return a dictionary of VideoOutputProxy objects for all tabs."""
        return [_make_voutput_proxy(voutput) for voutput in self.__workspace.outputs_manager.voutputs]

    @property
    def current_voutput(self) -> VideoOutputProxy:
        """Return the VideoOutput for the currently selected tab."""
        if voutput := self.__workspace.outputs_manager.current_voutput:
            return _make_voutput_proxy(voutput)

        # This shouldn't happen
        raise NotImplementedError

    # PRIVATE API
    @property
    def _settings_store(self) -> _PluginSettingsStore:
        if self.__settings_store is None:
            self.__settings_store = _PluginSettingsStore(self.__workspace)
        return self.__settings_store

    @run_in_loop(return_future=False)
    def _is_truly_visible(self, plugin: WidgetPluginBase[Any, Any]) -> bool:
        # Check if this plugin is truly visible to the user.

        # This accounts for:
        # - Widgets in a QTabWidget that are not the current tab
        # - Widgets in a QDockWidget that is tabified and not visible
        # - Widgets in a QSplitter panel that is collapsed (size=0)
        if not plugin.isVisible():
            return False

        widget: QObject | None = plugin

        while widget:
            parent = widget.parent()

            # Plugin is not the current tab (compare against self, not intermediate widget)
            if isinstance(parent, QTabWidget) and parent.currentWidget() is not plugin:
                return False

            # Dock widget is tabified and not visible
            if isinstance(parent, QDockWidget) and parent.visibleRegion().isEmpty():
                return False

            # Check if our panel in the splitter is collapsed
            if (
                isinstance(parent, QSplitter)
                and isinstance(widget, QWidget)
                and (idx := parent.indexOf(widget)) >= 0
                and parent.sizes()[idx] == 0
            ):
                return False

            widget = parent

        return True

    def _register_plugin_nodes_to_buffer(self) -> None:
        # Register visible plugin nodes with the buffer for pre-fetching during playback.
        from .api import PluginGraphicsView

        for plugin in self.__workspace.plugins:
            if not self._is_truly_visible(plugin):
                continue

            for view in self.__workspace.loop.from_thread(plugin.findChildren, PluginGraphicsView).result():
                if view.current_tab in view.outputs and self.__workspace.playback.state.buffer:
                    self.__workspace.playback.state.buffer.register_plugin_node(
                        plugin.identifier, view.outputs[view.current_tab]
                    )

    def _on_current_voutput_changed(self, refresh: bool = False) -> None:
        # Notify all visible plugin views of output change.
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                self._init_plugin(plugin, refresh)

    def _init_plugin(self, plugin: WidgetPluginBase[Any, Any], refresh: bool = False) -> None:
        # Initialize plugin for the current output and render initial frame if needed.
        from .api import PluginGraphicsView

        if not self._is_truly_visible(plugin):
            return

        try:
            plugin.on_current_voutput_changed(
                self.current_voutput,
                self.__workspace.outputs_manager.current_video_index,
            )
        except Exception:
            logger.exception("on_current_voutput_changed: Failed to initialize plugin %r", plugin)
            return

        try:
            plugin.on_current_frame_changed(self.__workspace.playback.state.current_frame)
        except Exception:
            logger.exception("on_current_frame_changed: Failed to initialize plugin %r", plugin)
            return

        for view in self.__workspace.loop.from_thread(plugin.findChildren, PluginGraphicsView).result():
            try:
                self._init_view(view, plugin, refresh)
            except Exception:
                logger.exception("Failed to initialize view %r", view)

    @run_in_loop(return_future=False)
    def _find_plugin_for_widget(self, widget: QWidget) -> WidgetPluginBase[Any, Any] | None:
        from .api import WidgetPluginBase

        current: QObject | None = widget.parent()

        while current is not None:
            if isinstance(current, WidgetPluginBase):
                return current
            current = current.parent()
        logger.warning("Could not find plugin for widget %r", widget)
        return None

    def _init_view(
        self, view: PluginGraphicsView, plugin: WidgetPluginBase[Any, Any] | None = None, refresh: bool = False
    ) -> None:
        if plugin is None and (plugin := self._find_plugin_for_widget(view)) is None:
            return

        if not self._is_truly_visible(plugin):
            return

        tab_index = self.__workspace.outputs_manager.current_video_index
        current_frame = self.__workspace.playback.state.current_frame

        # Detect if we are actually changing tabs or forcing a refresh
        image_changed = view.current_tab != tab_index or view.last_frame != current_frame
        view.current_tab = tab_index
        view.last_frame = current_frame

        logger.debug("Initializing view: %s, tab=%d (changed=%s), refresh=%s", view, tab_index, image_changed, refresh)

        if refresh:
            view.outputs.clear()

        if tab_index not in view.outputs:
            with self.__workspace.env.use():
                node = view.get_node(self.current_voutput.vs_output.clip)
                packed = self.__workspace.outputs_manager.packer.pack_clip(node)
                view.outputs[tab_index] = packed
                logger.debug("Created output node for tab %d", tab_index)

        with self.__workspace.env.use():
            view.on_current_voutput_changed(self.current_voutput, tab_index)

        if view.pixmap_item.pixmap().isNull() or image_changed or refresh:
            with self.__workspace.env.use(), view.outputs[tab_index].get_frame(current_frame) as frame:
                logger.debug("Rendering initial frame %d for view", current_frame)
                view.on_current_frame_changed(current_frame, frame)

        view.update_scene_rect()
        view.set_autofit(view.autofit)

    def _on_current_frame_changed(self, n: int, plugin_frames: dict[str, vs.VideoFrame] | None = None) -> None:
        # Notify plugins of frame change.
        # If plugin_frames is provided, uses pre-fetched frames.
        # Otherwise, fetches frames synchronously for each plugin view.
        from .api import PluginGraphicsView

        for plugin in self.__workspace.plugins:
            if not self._is_truly_visible(plugin):
                continue

            plugin.on_current_frame_changed(n)

            for view in self.__workspace.loop.from_thread(plugin.findChildren, PluginGraphicsView).result():
                if view.current_tab == -1 or view.current_tab not in view.outputs:
                    continue

                # Get pre-fetched frame or fall back to sync request
                if plugin_frames and plugin.identifier in plugin_frames:
                    view.on_current_frame_changed(n, plugin_frames[plugin.identifier])
                else:
                    with self.__workspace.env.use(), view.outputs[view.current_tab].get_frame(n) as frame:
                        view.on_current_frame_changed(n, frame)

    def _get_cached_proxy_settings(self, plugin: _PluginBase[Any, Any], scope: str) -> Any:
        model = self._settings_store.get(plugin, scope)

        if model is None:
            return model

        return _SettingsProxy(model, lambda k, v: self._settings_store.update(plugin, scope, **{k: v}))

    def _update_settings(self, plugin: _PluginBase[Any, Any], scope: str, **updates: Any) -> None:
        self._settings_store.update(plugin, scope, **updates)

    @run_in_loop(return_future=False)
    def _on_playback_started(self) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_playback_started()

    # Already run in loop from caller
    def _on_playback_stopped(self) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_playback_stopped()

    def _on_view_context_menu(self, event: QContextMenuEvent) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_view_context_menu(event)

    def _on_view_mouse_moved(self, event: QMouseEvent) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_view_mouse_moved(event)

    def _on_view_mouse_pressed(self, event: QMouseEvent) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_view_mouse_pressed(event)

    def _on_view_mouse_released(self, event: QMouseEvent) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_view_mouse_released(event)

    def _on_view_rect_selection_changed(self, rect: QRect) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_view_rect_selection_changed(rect)

    def _on_view_rect_selection_finished(self, rect: QRect) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_view_rect_selection_finished(rect)

    def _on_view_key_press(self, event: QKeyEvent) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_view_key_press(event)

    def _on_view_key_release(self, event: QKeyEvent) -> None:
        for plugin in self.__workspace.plugins:
            if self._is_truly_visible(plugin):
                plugin.on_view_key_release(event)

    def _on_global_settings_changed(self) -> None:
        self._settings_store.invalidate("global")
        self._settings_store.invalidate("local")
        self.globalSettingsChanged.emit()

    def _on_local_settings_changed(self, path: str) -> None:
        self._settings_store.invalidate("local")
        self.localSettingsChanged.emit(path)


class _PluginBaseMeta(ObjectType):
    global_settings_model: type[BaseModel] | None
    local_settings_model: type[BaseModel] | None

    def __new__[MetaSelf: _PluginBaseMeta](  # noqa: PYI019
        mcls: type[MetaSelf],
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        /,
        **kwargs: Any,
    ) -> MetaSelf:
        cls = super().__new__(mcls, name, bases, namespace, **kwargs)

        # Skip processing for base classes that set __plugin_base__ = True
        if namespace.get("__plugin_base__"):
            return cls

        # WidgetPluginBase and NodeProcessor are now defined so it's safe to import them
        from .api import NodeProcessor, WidgetPluginBase

        for base in get_original_bases(cls):
            if not (origin := get_origin(base)):
                continue

            args = get_args(base)

            if issubclass(origin, WidgetPluginBase):
                scope = ["global", "local"]
            elif issubclass(origin, NodeProcessor):
                scope = [None, "global", "local"]
            else:
                continue

            for n, arg in zip_longest(scope, args, fillvalue=None):
                if arg is None or not isinstance(arg, type):
                    setattr(cls, f"{n}_settings_model", None)
                    continue

                origin = get_origin(arg)
                is_basemodel = (origin is None and issubclass(arg, BaseModel)) or (
                    origin is not None and issubclass(origin, BaseModel)
                )
                setattr(cls, f"{n}_settings_model", arg if is_basemodel else None)
            break
        else:
            cls.global_settings_model, cls.local_settings_model = None, None

        return cls


class _GraphicsViewProxy(QObject):
    def __init__(self, workspace: LoaderWorkspace[Any], view: GraphicsView) -> None:
        super().__init__()
        self.__workspace = workspace
        self.__view = view


class _ViewportProxy(QObject):
    def __init__(self, workspace: LoaderWorkspace[Any], viewport: QWidget) -> None:
        super().__init__()
        self.__workspace = workspace
        self.__viewport = viewport
        self.__cursor_reset_conn: QMetaObject.Connection | None = None


class _TimelineProxy(QObject):
    def __init__(self, workspace: LoaderWorkspace[Any], timeline: Timeline) -> None:
        super().__init__()
        self.__workspace = workspace
        self.__timeline = timeline


class _PlaybackProxy(QObject):
    def __init__(self, workspace: LoaderWorkspace[Any], playback_manager: PlaybackManager) -> None:
        super().__init__()
        self.__workspace = workspace
        self.__playback_manager = playback_manager


class _PluginSecrets:
    def __init__(self, plugin: _PluginBase[Any, Any]) -> None:
        self.__namespace = plugin.identifier

    def __getattr__(self, name: str) -> Any:
        if name in dir(SecretsManager) and not name.startswith("_"):
            return lambda *args, **kwargs: getattr(SecretsManager, name)(self.__namespace, *args, **kwargs)

        raise AttributeError



from .api import (
    AudioOutputProxy,
    GraphicsViewProxy,
    LocalSettingsModel,
    NodeProcessor,
    PluginAPI,
    PluginGraphicsView,
    PluginSecrets,
    PluginSettings,
    VideoOutputProxy,
    WidgetPluginBase,
)
from .specs import hookimpl

__all__ = [
    "AudioOutputProxy",
    "GraphicsViewProxy",
    "LocalSettingsModel",
    "NodeProcessor",
    "PluginAPI",
    "PluginGraphicsView",
    "PluginSecrets",
    "PluginSettings",
    "VideoOutputProxy",
    "WidgetPluginBase",
    "hookimpl",
]



from __future__ import annotations

from bisect import bisect_left, bisect_right
from concurrent.futures import Future
from enum import StrEnum
from functools import cache
from itertools import count
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Self

import pluggy
from jetpytools import cachedproperty, flatten, to_arr
from pydantic import BaseModel, Field
from PySide6.QtCore import QPoint, Qt
from PySide6.QtGui import QAction, QKeySequence
from PySide6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QFileDialog,
    QHeaderView,
    QMenu,
    QSplitter,
    QTableView,
    QToolBar,
    QVBoxLayout,
    QWidget,
)

from vsview.api import (
    ActionDefinition,
    Frame,
    IconName,
    IconReloadMixin,
    PluginAPI,
    Time,
    VideoOutputProxy,
    WidgetPluginBase,
    run_in_background,
    run_in_loop,
)

from . import specs
from .models import RangeFrame, RangeTime, SceneRow, UnifiedRange
from .parsers import internal_parsers
from .serializer import internal_serializers
from .ui import Col, RangeCol, RangeTableDelegate, RangeTableModel, SceneTableDelegate, SceneTableModel
from .utils import ColorGenerator, monkey_patch_parser

if TYPE_CHECKING:
    from .api import Parser, Serializer

logger = getLogger(__name__)


PLUGIN_IDENTIFIER = "jet_vsview_scening"
PLUGIN_DISPLAY_NAME = "Scening"


manager = pluggy.PluginManager("vsview.scening")


@cache
def load_plugins() -> None:
    manager.add_hookspecs(specs)
    n = manager.load_setuptools_entrypoints("vsview.scening")
    logger.debug("Loaded %d plugins", n)


class ShortcutDefinition(StrEnum):
    definition: ActionDefinition

    REMOVE_SCENE = "remove_scene", "Remove selected scene", "Delete"

    TOGGLE_RANGE_START = "toggle_range_start", "Toggle Range Start", "Q"
    TOGGLE_RANGE_END = "toggle_range_end", "Toggle Range End", "W"
    VALIDATE_RANGE = "validate_range", "Validate selected range", "E"
    ADD_SINGLE_FRAME = "add_single_frame", "Add Single Frame", "R"
    REMOVE_RANGE = "remove_range", "Remove selected range", "Delete"

    SEEK_PREV_BOUND = "seek_prev_bound", "Seek to previous range boundary", "Ctrl+Left"
    SEEK_NEXT_BOUND = "seek_next_bound", "Seek to next range boundary", "Ctrl+Right"
    SEEK_PREV_RANGE = "seek_prev_range", "Seek to previous range", "Ctrl+Shift+Left"
    SEEK_NEXT_RANGE = "seek_next_range", "Seek to next range", "Ctrl+Shift+Right"

    def __new__(cls, value: str, label: str, default_key: str = "") -> Self:
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.definition = ActionDefinition(f"{PLUGIN_IDENTIFIER}.{value}", label, default_key)
        return obj


class LocalSettings(BaseModel):
    scenes: list[SceneRow] = Field(default_factory=list)


class SceningPlugin(WidgetPluginBase[None, LocalSettings], IconReloadMixin):
    identifier = PLUGIN_IDENTIFIER
    display_name = PLUGIN_DISPLAY_NAME

    shortcuts = tuple(s.definition for s in ShortcutDefinition)

    def __init__(self, parent: QWidget, api: PluginAPI) -> None:
        super().__init__(parent, api)
        IconReloadMixin.__init__(self)

        self.output_map = dict[int, str]()
        self._pending_start: Frame | Time | None = None
        self._pending_end: Frame | Time | None = None

        self.setup_ui()
        self.setup_shortcuts()
        self.load_settings()

        self.register_icon_callback(self.on_reload_icon)
        self.api.register_on_destroy(self.init_load.cache_clear)

        self._update_action_labels()

    def setup_ui(self) -> None:
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.splitter = QSplitter(self, orientation=Qt.Orientation.Vertical)
        layout.addWidget(self.splitter)

        self.scenes_container = QWidget(self.splitter)
        scenes_layout = QVBoxLayout(self.scenes_container)
        scenes_layout.setContentsMargins(0, 0, 0, 0)
        scenes_layout.setSpacing(0)

        # Top toolbar
        toolbar = QToolBar(self.scenes_container, movable=False)
        scenes_layout.addWidget(toolbar)

        self.new_scene_btn = self.make_tool_button(IconName.PLUS, "Create a new scene", self.scenes_container)
        self.new_scene_btn.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.new_scene_btn.setText("Create new scene")
        self.new_scene_btn.clicked.connect(self.on_new_scene)
        toolbar.addWidget(self.new_scene_btn)

        self.import_scene_btn = self.make_tool_button(
            IconName.FILE_IMPORT,
            "Import a file suitable for scening",
            self.scenes_container,
        )
        self.import_scene_btn.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.import_scene_btn.setText("Import scene...")
        self.import_scene_btn.clicked.connect(self.on_import_scene)
        toolbar.addWidget(self.import_scene_btn)

        self.export_scene_btn = self.make_tool_button(
            IconName.FILE_EXPORT,
            "Export a scene to a supported format",
            self.scenes_container,
        )
        self.export_scene_btn.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.export_scene_btn.setText("Export scene...")
        self.export_scene_btn.setDisabled(True)
        self.export_scene_btn.clicked.connect(self.on_export_scene)
        toolbar.addWidget(self.export_scene_btn)

        # Scenes model + delegate
        self.scenes_model = SceneTableModel(self.output_map, self.scenes_container)
        self.scenes_model.scenesModified.connect(self._persist_scenes)
        self.scenes_model.sceneDisplayModified.connect(self._refresh_scene_on_timeline)
        self.scenes_model.sceneColorModified.connect(self._refresh_scene_on_timeline)
        self.scenes_model.sceneCheckOutputsModified.connect(self._refresh_scene_on_timeline)
        self.scenes_delegate = SceneTableDelegate(self.output_map, self.scenes_container)

        self.scenes_view = QTableView(self.scenes_container, showGrid=False)
        self.scenes_view.setAlternatingRowColors(False)
        self.scenes_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.scenes_view.setSortingEnabled(False)
        self.scenes_view.verticalHeader().setVisible(False)
        self.scenes_view.setModel(self.scenes_model)
        self.scenes_view.setItemDelegate(self.scenes_delegate)
        self.scenes_view.selectionModel().selectionChanged.connect(self.on_scene_selection_changed)

        # Scenes column sizing
        header = self.scenes_view.horizontalHeader()
        header.setSectionResizeMode(Col.COLOR, QHeaderView.ResizeMode.Fixed)
        header.resizeSection(Col.COLOR, 40)
        header.setSectionResizeMode(Col.NAME, QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(Col.OUTPUTS, QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(Col.DISPLAY, QHeaderView.ResizeMode.Fixed)
        header.resizeSection(Col.DISPLAY, 55)
        header.setSectionResizeMode(Col.DELETE, QHeaderView.ResizeMode.Fixed)
        header.resizeSection(Col.DELETE, 30)

        scenes_layout.addWidget(self.scenes_view)
        self.splitter.addWidget(self.scenes_container)

        self.range_container = QWidget(self.splitter)
        self.range_container.setDisabled(True)
        range_layout = QVBoxLayout(self.range_container)
        range_layout.setContentsMargins(0, 0, 0, 0)
        range_layout.setSpacing(0)

        # Range toolbar
        range_toolbar = QToolBar(
            self.range_container,
            movable=False,
            toolButtonStyle=Qt.ToolButtonStyle.ToolButtonTextBesideIcon,
        )
        range_layout.addWidget(range_toolbar)

        self.range_start_action = self.make_action(
            IconName.MARK_IN,
            "Toggle start of new range",
            self.range_container,
            checkable=True,
            icon_states=self.DEFAULT_ICON_STATES,
        )
        self.range_start_action.setIconText("Mark in")
        self.range_start_action.toggled.connect(self.on_range_start_toggle)

        self.range_end_action = self.make_action(
            IconName.MARK_OUT,
            "Toggle end of new range",
            self.range_container,
            checkable=True,
            icon_states=self.DEFAULT_ICON_STATES,
        )
        self.range_end_action.setIconText("Mark out")
        self.range_end_action.toggled.connect(self.on_range_end_toggle)

        self.add_range_action = self.make_action(
            IconName.SCENE_ADD,
            "Add current selected range",
            self.range_container,
            icon_states=self.DEFAULT_ICON_STATES,
        )
        self.add_range_action.setIconText("Add range")
        self.add_range_action.setDisabled(True)
        self.add_range_action.triggered.connect(self.on_add_range_triggered)

        self.add_frame_action = self.make_action(
            IconName.FRAME_ADD,
            "Add current selected frame as range without end boudary",
            self.range_container,
            icon_states=self.DEFAULT_ICON_STATES,
        )
        self.add_frame_action.setIconText("Add frame")
        self.add_frame_action.triggered.connect(self.on_add_frame_triggered)

        range_toolbar.addActions(
            [
                self.range_start_action,
                self.range_end_action,
                self.add_range_action,
                self.add_frame_action,
            ]
        )

        self.ranges_model = RangeTableModel(self.range_container, self.api)
        self.ranges_model.rangesModified.connect(self._persist_scenes)
        self.ranges_model.rangeDataModified.connect(self._refresh_range_on_timeline)
        self.ranges_model.modelReset.connect(self._update_ranges_header_width)
        self.ranges_model.rowsInserted.connect(self._update_ranges_header_width)
        self.ranges_model.rowsRemoved.connect(self._update_ranges_header_width)

        self.ranges_delegate = RangeTableDelegate(self.range_container)

        self.ranges_view = QTableView(self.range_container)
        self.ranges_view.setAlternatingRowColors(True)
        self.ranges_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.ranges_view.setSortingEnabled(True)
        self.ranges_view.verticalHeader().setVisible(True)
        self.ranges_view.setModel(self.ranges_model)
        self.ranges_view.setItemDelegate(self.ranges_delegate)
        self.ranges_view.selectionModel().selectionChanged.connect(self.on_range_selection_changed)

        # Context menu & Ctrl+C
        self.ranges_view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.ranges_view.customContextMenuRequested.connect(self._on_ranges_context_menu)

        self.copy_frames_action = QAction(
            "Copy frames",
            self.ranges_view,
            shortcut=QKeySequence(QKeySequence.StandardKey.Copy),
            shortcutContext=Qt.ShortcutContext.WidgetShortcut,
        )
        self.copy_frames_action.triggered.connect(self._copy_ranges_frames)
        self.ranges_view.addAction(self.copy_frames_action)

        r_header = self.ranges_view.horizontalHeader()
        r_header.setSectionResizeMode(RangeCol.START_FRAME, QHeaderView.ResizeMode.Interactive)
        r_header.setSectionResizeMode(RangeCol.END_FRAME, QHeaderView.ResizeMode.Interactive)
        r_header.setSectionResizeMode(RangeCol.START_TIME, QHeaderView.ResizeMode.Interactive)
        r_header.setSectionResizeMode(RangeCol.END_TIME, QHeaderView.ResizeMode.Interactive)
        r_header.setSectionResizeMode(RangeCol.LABEL, QHeaderView.ResizeMode.Stretch)
        r_header.resizeSection(RangeCol.START_FRAME, 72)
        r_header.resizeSection(RangeCol.END_FRAME, 72)
        r_header.resizeSection(RangeCol.START_TIME, 88)
        r_header.resizeSection(RangeCol.END_TIME, 88)
        r_header.setMinimumSectionSize(67)

        v_header = self.ranges_view.verticalHeader()
        v_header.setSectionResizeMode(QHeaderView.ResizeMode.Fixed)
        v_header.setDefaultSectionSize(24)

        range_layout.addWidget(self.ranges_view)
        self.splitter.addWidget(self.range_container)

    def setup_shortcuts(self) -> None:
        self.api.register_shortcut(
            ShortcutDefinition.REMOVE_SCENE.definition,
            self.on_remove_scene_triggered,
            self.scenes_view,
            context=Qt.ShortcutContext.WidgetShortcut,
        )
        self.api.register_action(
            ShortcutDefinition.TOGGLE_RANGE_START.definition,
            self.range_start_action,
            context=Qt.ShortcutContext.WindowShortcut,
        )
        self.api.register_action(
            ShortcutDefinition.TOGGLE_RANGE_END.definition,
            self.range_end_action,
            context=Qt.ShortcutContext.WindowShortcut,
        )
        self.api.register_action(
            ShortcutDefinition.VALIDATE_RANGE.definition,
            self.add_range_action,
            context=Qt.ShortcutContext.WindowShortcut,
        )
        self.api.register_action(
            ShortcutDefinition.ADD_SINGLE_FRAME.definition,
            self.add_frame_action,
            context=Qt.ShortcutContext.WindowShortcut,
        )
        self.api.register_shortcut(
            ShortcutDefinition.REMOVE_RANGE.definition,
            self.on_remove_range_triggered,
            self.ranges_view,
            context=Qt.ShortcutContext.WidgetShortcut,
        )

        self.api.register_shortcut(
            ShortcutDefinition.SEEK_PREV_BOUND.definition,
            self.on_seek_prev_bound,
            self.ranges_view,
            context=Qt.ShortcutContext.WindowShortcut,
        )
        self.api.register_shortcut(
            ShortcutDefinition.SEEK_NEXT_BOUND.definition,
            self.on_seek_next_bound,
            self.ranges_view,
            context=Qt.ShortcutContext.WindowShortcut,
        )
        self.api.register_shortcut(
            ShortcutDefinition.SEEK_PREV_RANGE.definition,
            self.on_seek_prev_range,
            self.ranges_view,
            context=Qt.ShortcutContext.WindowShortcut,
        )
        self.api.register_shortcut(
            ShortcutDefinition.SEEK_NEXT_RANGE.definition,
            self.on_seek_next_range,
            self.ranges_view,
            context=Qt.ShortcutContext.WindowShortcut,
        )

    def load_settings(self) -> None:
        self._color_gen = ColorGenerator(None)

        # Set the next color based on the last scene
        if scenes := self.settings.local_.scenes:
            next(self._color_gen)
            self._color_gen.send(scenes[-1].color)

        self._counter = count(len(scenes) + 1)

        # Load scenes from settings
        for scene in self.settings.local_.scenes:
            self.scenes_model.add_scene(scene, emit_signal=False)

    @cache
    def init_load(self) -> None:
        self.output_map.clear()
        self.output_map.update((out.vs_index, out.vs_name) for out in self.api.voutputs)

    def on_current_voutput_changed(self, voutput: VideoOutputProxy, tab_index: int) -> None:
        self.init_load()
        cachedproperty.clear_cache(self.ranges_model)

        self.on_scene_selection_changed()

        return super().on_current_voutput_changed(voutput, tab_index)

    def on_reload_icon(self) -> None:
        cachedproperty.clear_cache(self.scenes_delegate)
        self.scenes_view.viewport().update()
        self._update_action_labels()

    def on_new_scene(self) -> None:
        new_scene = SceneRow(color=next(self._color_gen), name=f"New Scene {str(next(self._counter)).zfill(2)}")

        idx = self.scenes_model.add_scene(new_scene)
        self.scenes_view.scrollTo(idx)
        self.scenes_view.setCurrentIndex(idx)

    def on_import_scene(self) -> None:
        load_plugins()

        parsers: list[Parser] = internal_parsers + list(flatten(manager.hook.vsview_scening_register_parser()))

        filters = {f"{p.filter.label} (*.{' *.'.join(to_arr(p.filter.suffix))})": p for p in parsers}
        files, selected_filter = QFileDialog.getOpenFileNames(
            self,
            "Import scene file(s)",
            filter=";;".join(sorted(filters)),
        )

        if not files:
            logger.info("No file selected")
            return

        fscenes = self.parse_imported_files(files, filters[selected_filter])

        @run_in_loop
        def on_completed(f: Future[list[SceneRow]]) -> None:
            if not f.exception():
                self.scenes_model.add_scene(f.result())
                self.api.statusMessage.emit(f"Importing {', '.join(files)} completed")

        fscenes.add_done_callback(on_completed)

    @run_in_background(name="ParseImportedFiles")
    def parse_imported_files(self, files: list[str], parser: Parser) -> list[SceneRow]:
        src_fps = self.api.current_voutput.vs_output.clip.fps

        scenes = list[SceneRow]()

        with monkey_patch_parser(parser, self._color_gen):
            for file in files:
                pfile = Path(file)
                try:
                    with pfile.open("rb") as f:
                        parsed = parser.parse(f, pfile.stem, src_fps)
                        scenes.extend([parsed] if isinstance(parsed, SceneRow) else parsed)
                except Exception:
                    logger.exception("Error parsing file: %s", file)

        return scenes

    def on_export_scene(self) -> None:
        load_plugins()

        serializers: list[Serializer] = internal_serializers + list(
            flatten(manager.hook.vsview_scening_register_serializer())
        )

        filters = {f"{p.filter.label} (*.{' *.'.join(to_arr(p.filter.suffix))})": p for p in serializers}
        file, selected_filter = QFileDialog.getSaveFileName(
            self,
            "Export scene file",
            filter=";;".join(sorted(filters)),
        )

        if not file:
            logger.info("Export aborted")
            return

        f = self.serialize_exported_scene(file, filters[selected_filter])

        @run_in_loop
        def on_completed(f: Future[None]) -> None:
            if not f.exception():
                self.api.statusMessage.emit(f"Exporting {file} completed")
                return

        f.add_done_callback(on_completed)

    @run_in_background(name="SerializeExportedCcene")
    def serialize_exported_scene(self, file: str, serializer: Serializer) -> None:
        scene: SceneRow = self.scenes_view.selectionModel().selectedRows()[0].data(self.scenes_model.SceneRowRole)

        v = self.api.current_voutput

        try:
            with Path(file).open("wb") as f:
                serializer.serialize(f, (UnifiedRange(r, v.frame_to_time, v.time_to_frame) for r in scene.ranges))
        except Exception:
            logger.exception("Error exporting file: %s", file)

    def on_remove_scene_triggered(self) -> None:
        if not (selected_indexes := self.scenes_view.selectionModel().selectedRows()):
            return

        rows = sorted((idx.row() for idx in selected_indexes), reverse=True)

        self.scenes_model.remove_scene(rows)

    def on_scene_selection_changed(self) -> None:
        if selected_indexes := self.scenes_view.selectionModel().selectedRows():
            self.range_container.setEnabled(True)
            if len(selected_indexes) == 1:
                self.export_scene_btn.setEnabled(True)
            else:
                self.export_scene_btn.setEnabled(False)

            all_scenes = list[SceneRow]()

            for index in selected_indexes:
                all_scenes.append(index.data(self.scenes_model.SceneRowRole))

            self.ranges_model.set_scenes(all_scenes)

            # Add displayable scenes to the timeline
            for scene in all_scenes:
                self._refresh_scene_on_timeline(scene, update=False)

            # Remove old references
            self.api.timeline.clear_notches(
                (scene.notch_id for scene in set(self.settings.local_.scenes).difference(all_scenes)),
                update=True,
            )
        else:
            self.range_container.setDisabled(True)
            self.export_scene_btn.setDisabled(True)
            self.range_start_action.setChecked(False)

            self.api.timeline.clear_notches(scene.notch_id for scene in self.settings.local_.scenes)

    def on_range_selection_changed(self) -> None:
        selected = self.ranges_view.selectionModel().selectedRows()

        if len(selected) == 1:
            start, _ = self.ranges_model.data(selected[0], self.ranges_model.RangeRole).as_frames(
                self.api.current_voutput
            )
            self.api.playback.seek(start)

    def on_range_start_toggle(self, checked: bool) -> None:
        if checked:
            self._pending_start = self.api.current_frame if self.api.timeline.mode == "frame" else self.api.current_time

            if self.range_end_action.isChecked():
                self.add_range_action.setEnabled(True)
        else:
            self._pending_start = None
            self.add_range_action.setDisabled(True)

    def on_range_end_toggle(self, checked: bool) -> None:
        if checked:
            self._pending_end = self.api.current_frame if self.api.timeline.mode == "frame" else self.api.current_time

            if self.range_start_action.isChecked():
                self.add_range_action.setEnabled(True)
        else:
            self._pending_end = None
            self.add_range_action.setDisabled(True)

    def on_add_range_triggered(self) -> None:
        selected_indexes = self.scenes_view.selectionModel().selectedRows()

        # Shouldn't happen
        if not selected_indexes or self._pending_start is None or self._pending_end is None:
            raise NotImplementedError

        index = next(reversed(selected_indexes))
        scene: SceneRow = index.data(self.scenes_model.SceneRowRole)
        self._add_range(self._pending_start, self._pending_end, scene)

    def on_add_frame_triggered(self) -> None:
        selected_indexes = self.scenes_view.selectionModel().selectedRows()

        # Shouldn't happen
        if not selected_indexes:
            raise NotImplementedError

        index = next(reversed(selected_indexes))
        scene: SceneRow = index.data(self.scenes_model.SceneRowRole)
        current = self.api.current_frame if self.api.timeline.mode == "frame" else self.api.current_time
        self._add_range(current, current, scene)

    def on_remove_range_triggered(self) -> None:
        if not (selected_indexes := self.ranges_view.selectionModel().selectedRows()):
            return

        # Sort by row descending to avoid index shifts
        for idx in sorted(selected_indexes, key=lambda idx: idx.row(), reverse=True):
            r = self.ranges_model.data(idx, self.ranges_model.RangeRole)
            scene = self.ranges_model.data(idx, self.ranges_model.SceneRowRole)

            self.ranges_model.remove_range(idx)
            self.api.timeline.discard_notch(scene.notch_id, [r.to_tuple()], r.id, update=False)

        self.api.timeline.update()

    def on_seek_prev_bound(self) -> None:
        self._seek_to_neighbor(self._get_visible_range_boundaries(), forward=False)

    def on_seek_next_bound(self) -> None:
        self._seek_to_neighbor(self._get_visible_range_boundaries(), forward=True)

    def on_seek_prev_range(self) -> None:
        self._seek_to_neighbor(self._get_visible_range_boundaries(starts_only=True), forward=False)

    def on_seek_next_range(self) -> None:
        self._seek_to_neighbor(self._get_visible_range_boundaries(starts_only=True), forward=True)

    def _add_range(self, start: Frame | Time, end: Frame | Time, scene: SceneRow) -> None:
        r: RangeFrame | RangeTime

        if isinstance(start, Frame):
            if not isinstance(end, Frame):
                end = self.api.current_voutput.time_to_frame(end)
            s, e = sorted([start, end])
            r = RangeFrame(start=s, end=e)
        elif isinstance(start, Time):
            if not isinstance(end, Time):
                end = self.api.current_voutput.frame_to_time(end)
            s, e = sorted([start, end])
            r = RangeTime(start=s, end=e)

        self.ranges_model.add_range(r, scene)

        self.ranges_view.scrollToBottom()
        self.range_start_action.setChecked(False)
        self.range_end_action.setChecked(False)

        self.api.timeline.add_notch(scene.notch_id, [r.to_tuple()], scene.color, r.label, r.id)

    def _persist_scenes(self) -> None:
        scenes = self.scenes_model.scenes.copy()
        self.settings.local_.scenes = scenes

        if scenes:
            self._color_gen.send(scenes[-1].color)

    def _refresh_scene_on_timeline(self, scene: SceneRow, *, update: bool = True) -> None:
        self.api.timeline.clear_notches(scene.notch_id, update=False)

        if scene.display and (not scene.checked_outputs or self.api.current_voutput.vs_index in scene.checked_outputs):
            for r in scene.ranges:
                self.api.timeline.add_notch(scene.notch_id, [r.to_tuple()], scene.color, r.label, r.id, update=False)

        if update:
            self.api.timeline.update()

    def _refresh_range_on_timeline(self, r: RangeFrame | RangeTime, scene: SceneRow, *, update: bool = True) -> None:
        self.api.timeline.discard_notch(scene.notch_id, [r.to_tuple()], r.id, update=False)
        self.api.timeline.add_notch(scene.notch_id, [r.to_tuple()], scene.color, r.label, r.id, update=False)

        if update:
            self.api.timeline.update()

    def _selected_ranges(self) -> list[RangeFrame | RangeTime]:
        return [idx.data(self.ranges_model.RangeRole) for idx in self.ranges_view.selectionModel().selectedRows()]

    def _copy_ranges_frames(self) -> None:
        v = self.api.current_voutput
        text = ", ".join(str(r.as_frames(v)) for r in self._selected_ranges())

        QApplication.clipboard().setText(f"[{text}]")

    def _copy_ranges_start_frames(self) -> None:
        v = self.api.current_voutput
        text = ", ".join(str(r.as_frames(v)[0]) for r in self._selected_ranges())

        QApplication.clipboard().setText(f"[{text}]")

    def _copy_ranges_end_frames(self) -> None:
        v = self.api.current_voutput
        text = ", ".join(str(r.as_frames(v)[1]) for r in self._selected_ranges())

        QApplication.clipboard().setText(f"[{text}]")

    def _copy_ranges_timestamps(self) -> None:
        v = self.api.current_voutput
        parts = list[str]()

        for r in self._selected_ranges():
            s, e = r.as_times(v)
            parts.append(f'("{s.to_ts()}", "{e.to_ts()}")')

        QApplication.clipboard().setText(f"[{', '.join(parts)}]")

    def _copy_ranges_start_timestamps(self) -> None:
        v = self.api.current_voutput
        text = ", ".join(f'"{r.as_times(v)[0].to_ts()}"' for r in self._selected_ranges())

        QApplication.clipboard().setText(f"[{text}]")

    def _copy_ranges_end_timestamps(self) -> None:
        v = self.api.current_voutput
        text = ", ".join(f'"{r.as_times(v)[1].to_ts()}"' for r in self._selected_ranges())

        QApplication.clipboard().setText(f"[{text}]")

    def _copy_ranges_labels(self) -> None:
        text = ", ".join(f'"{r.label}"' for r in self._selected_ranges())

        QApplication.clipboard().setText(f"[{text}]")

    def _on_ranges_context_menu(self, pos: QPoint) -> None:
        if not self.ranges_view.selectionModel().selectedRows():
            return

        menu = QMenu(self.ranges_view)
        menu.addAction(self.copy_frames_action)
        menu.addAction("Copy start frames only", self._copy_ranges_start_frames)
        menu.addAction("Copy end frames only", self._copy_ranges_end_frames)
        menu.addSeparator()

        menu.addAction("Copy timestamps", self._copy_ranges_timestamps)
        menu.addAction("Copy start timestamps only", self._copy_ranges_start_timestamps)
        menu.addAction("Copy end timestamps only", self._copy_ranges_end_timestamps)
        menu.addSeparator()

        menu.addAction("Copy labels", self._copy_ranges_labels)

        menu.exec(self.ranges_view.viewport().mapToGlobal(pos))
        menu.deleteLater()

    def _get_visible_range_boundaries(self, starts_only: bool = False) -> list[int]:
        v = self.api.current_voutput
        points = set[int]()

        for r, _ in self.ranges_model.ranges:
            start, end = r.as_frames(v)
            points.add(start)

            if not starts_only:
                points.add(end)

        return sorted(points)

    def _seek_to_neighbor(self, points: list[int], forward: bool) -> None:
        if not points:
            return

        current = int(self.api.current_frame)

        if forward:
            idx = bisect_right(points, current)

            if idx < len(points):
                self.api.playback.seek(points[idx])
        else:
            idx = bisect_left(points, current)

            if idx > 0:
                self.api.playback.seek(points[idx - 1])

    def _update_ranges_header_width(self) -> None:
        rows = self.ranges_model.rowCount()
        width = self.ranges_view.verticalHeader().fontMetrics().horizontalAdvance(str(rows)) + 12
        self.ranges_view.verticalHeader().setFixedWidth(max(width, 24))

    def _update_action_labels(self) -> None:
        def set_text(action: QAction, action_id: str, base_text: str) -> None:
            key = self.api.get_shortcut_label(action_id)
            action.setIconText(f"{base_text} ({key})" if key else base_text)

        set_text(self.range_start_action, ShortcutDefinition.TOGGLE_RANGE_START.definition, "Mark in")
        set_text(self.range_end_action, ShortcutDefinition.TOGGLE_RANGE_END.definition, "Mark out")
        set_text(self.add_range_action, ShortcutDefinition.VALIDATE_RANGE.definition, "Add range")
        set_text(self.add_frame_action, ShortcutDefinition.ADD_SINGLE_FRAME.definition, "Add frame")



from __future__ import annotations

from abc import abstractmethod
from collections import deque
from collections.abc import Callable, Iterator
from concurrent.futures import Future
from contextlib import contextmanager
from functools import partial
from logging import DEBUG, getLogger
from pathlib import Path
from threading import Lock
from types import ModuleType
from typing import Any, ClassVar, Literal, assert_never

from jetpytools import clamp, fallback
from PySide6.QtCore import QSignalBlocker, Qt, Signal
from PySide6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QPushButton,
    QStackedWidget,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)
from vsengine.policy import ManagedEnvironment
from vsengine.vpy import ExecutionError, Script, load_code, load_script

from ...vsenv import gc_collect, run_in_background, run_in_loop, unset_environment
from ..outputs import AudioOutput, OutputsManager, VideoOutput
from ..plugins.api import PluginAPI, WidgetPluginBase
from ..plugins.manager import PluginManager
from ..settings import ActionID, ShortcutManager
from ..views import PluginDock, PluginSplitter
from ..views.components import CustomLoadingPage, DockButton
from ..views.timeline import Frame, TimelineControlBar
from .base import BaseWorkspace
from .playback import PlaybackManager
from .tab_manager import PlayHeadToolButton, TabManager

loader_lock = Lock()
logger = getLogger(__name__)


class LoaderWorkspace[T](BaseWorkspace):
    """A workspace that supports loading content."""

    content: T
    """The content being loaded."""

    # Status bar signals
    statusLoadingStarted = Signal(str)  # message
    statusLoadingFinished = Signal(str)  # completed message
    statusLoadingErrored = Signal(str)  # error message
    statusOutputChanged = Signal(object)  # OutputInfo dataclass

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.stack = QStackedWidget(self)
        self.current_layout.addWidget(self.stack)

        # API & plugins
        self.api = PluginAPI(self)
        self.cbs_on_destroy = list[Callable[[], Any]]()
        self.plugins = list[WidgetPluginBase[Any, Any]]()
        self.docks = list[PluginDock]()
        self.plugins_loaded = False

        # Empty State
        self.empty_page = QWidget(self)
        self.empty_layout = QVBoxLayout(self.empty_page)
        self.empty_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.load_btn = QPushButton(f"Load {self.title}")
        self.load_btn.setFixedSize(200, 50)
        self.empty_layout.addWidget(self.load_btn)
        self.stack.addWidget(self.empty_page)

        # Error State (failed content with reload option)
        self.error_page = QWidget(self)
        self.error_layout = QHBoxLayout(self.error_page)
        self.error_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.reload_btn = QPushButton(f"Reload {self.title}")
        self.reload_btn.setFixedSize(200, 50)
        self.reload_btn.clicked.connect(self._on_reload_failed)
        self.error_layout.addWidget(self.reload_btn)
        self.error_load_btn = QPushButton(f"Load {self.title}")
        self.error_load_btn.setFixedSize(200, 50)
        self.error_layout.addWidget(self.error_load_btn)
        self.stack.addWidget(self.error_page)

        # Loading State
        self.loading_page = CustomLoadingPage(self)
        self.stack.addWidget(self.loading_page)

        # Loaded State
        self.loaded_page = QWidget(self)
        self.loaded_layout = QVBoxLayout(self.loaded_page)
        self.loaded_layout.setContentsMargins(0, 0, 0, 0)
        self.loaded_layout.setSpacing(0)

        # Horizontal container for toggle button and main content
        self.content_area = QWidget(self.loaded_page)
        self.content_layout = QHBoxLayout(self.content_area)
        self.content_layout.setContentsMargins(0, 0, 0, 0)
        self.content_layout.setSpacing(0)

        # Left dock toggle button styled as a splitter handle
        self.dock_toggle_btn = DockButton(self.content_area)
        self.dock_toggle_btn.raise_()
        self.dock_toggle_btn.clicked.connect(self._on_dock_toggle)
        self.content_layout.addWidget(self.dock_toggle_btn, alignment=Qt.AlignmentFlag.AlignRight)

        # Embedded QMainWindow for dock widget support in the view area
        self.dock_container = QMainWindow(self.content_area)
        self.dock_container.setWindowFlags(Qt.WindowType.Widget)
        for area in (
            Qt.DockWidgetArea.LeftDockWidgetArea,
            Qt.DockWidgetArea.RightDockWidgetArea,
            Qt.DockWidgetArea.TopDockWidgetArea,
            Qt.DockWidgetArea.BottomDockWidgetArea,
        ):
            self.dock_container.setTabPosition(area, QTabWidget.TabPosition.North)

        self.plugin_splitter = PluginSplitter(self.dock_container)

        # Video Area (Tabs)
        self.tab_manager = TabManager(self.plugin_splitter, self.api)
        self.tab_manager.tabChanged.connect(self._on_tab_changed)
        self.tab_manager.sarTransformed.connect(lambda _: self._emit_output_info())
        self.plugin_splitter.insert_main_widget(self.tab_manager)

        # Connect plugin visibility signals
        self.tab_manager.toggle_toolpanel_btn.toggled.connect(self.plugin_splitter.toggle_right_panel)
        self.plugin_splitter.rightPanelVisibilityChanged.connect(self._sync_toolpanel_btn)
        self.plugin_splitter.rightPanelVisibilityChanged.connect(self._on_splitter_visibility_changed)
        self.plugin_splitter.pluginTabChanged.connect(self._on_splitter_tab_changed)

        self.dock_container.setCentralWidget(self.plugin_splitter)
        self.content_layout.addWidget(self.dock_container)

        self.loaded_layout.addWidget(self.content_area)

        # Timeline and Playback Controls
        self.tbar = TimelineControlBar(self)
        self.loaded_layout.addWidget(self.tbar)
        self.stack.addWidget(self.loaded_page)

        self.outputs_manager = OutputsManager()

        # PlaybackManager - handles video/audio playback logic
        self.playback = PlaybackManager(
            self.loop,
            lambda: self.env,
            self.api,
            self.outputs_manager,
            self.tab_manager,
            self.tbar,
            parent=self,
        )

        # Connect PlaybackManager signals to UI handlers
        self.playback.frameRendered.connect(lambda image, sar: self.tab_manager.update_current_view(image, sar=sar))
        self.playback.timelineCursorChanged.connect(self.update_timeline_cursor)
        self.playback.loadFailed.connect(self.clear_failed_load)

        # Audio control signals
        self.playback.audioOutputChanged.connect(
            lambda index: setattr(self.outputs_manager, "current_audio_index", index)
        )

        self._register_shortcuts()

    def _register_shortcuts(self) -> None:
        """Register workspace shortcuts with the shortcut manager."""
        sm = ShortcutManager()

        # Playback controls
        sm.register_shortcut(
            ActionID.PLAY_PAUSE,
            self.tbar.playback_container.play_pause_btn.click,
            self.loaded_page,
        )
        sm.register_shortcut(
            ActionID.SEEK_PREVIOUS_FRAME,
            self.tbar.playback_container.seek_1_back_btn.click,
            self.loaded_page,
        )
        sm.register_shortcut(
            ActionID.SEEK_NEXT_FRAME,
            self.tbar.playback_container.seek_1_fwd_btn.click,
            self.loaded_page,
        )
        sm.register_shortcut(
            ActionID.SEEK_N_FRAMES_BACK,
            self.tbar.playback_container.seek_n_back_btn.click,
            self.loaded_page,
        )
        sm.register_shortcut(
            ActionID.SEEK_N_FRAMES_FORWARD,
            self.tbar.playback_container.seek_n_fwd_btn.click,
            self.loaded_page,
        )
        sm.register_shortcut(
            ActionID.SEEK_FIRST_FRAME,
            lambda: self.playback.request_frame(0),
            self.loaded_page,
        )
        sm.register_shortcut(
            ActionID.SEEK_LAST_FRAME,
            lambda: self.playback.request_frame(0x7FFFFFFF),
            self.loaded_page,
        )
        sm.register_shortcut(ActionID.RELOAD, self.reload_content, self.loaded_page)
        sm.register_shortcut(ActionID.RELOAD, self.reload_btn.click, self.error_page)
        sm.register_shortcut(ActionID.COPY_CURRENT_FRAME, self._copy_current_frame_to_clipboard, self.loaded_page)
        sm.register_shortcut(ActionID.COPY_CURRENT_TIME, self._copy_current_time_to_clipboard, self.loaded_page)

        tab_actions = (
            ActionID.SWITCH_TAB_0,
            ActionID.SWITCH_TAB_1,
            ActionID.SWITCH_TAB_2,
            ActionID.SWITCH_TAB_3,
            ActionID.SWITCH_TAB_4,
            ActionID.SWITCH_TAB_5,
            ActionID.SWITCH_TAB_6,
            ActionID.SWITCH_TAB_7,
            ActionID.SWITCH_TAB_8,
            ActionID.SWITCH_TAB_9,
        )
        for i, action in enumerate(tab_actions):
            sm.register_shortcut(action, partial(self.tab_manager.switch_tab, i), self)

        sm.register_shortcut(ActionID.SWITCH_PREVIOUS_TAB, lambda: self.tab_manager.switch_tab(delta=-1), self)
        sm.register_shortcut(ActionID.SWITCH_NEXT_TAB, lambda: self.tab_manager.switch_tab(delta=1), self)

    def deleteLater(self) -> None:
        logger.debug(
            "%s(%r) deleteLater called, cleaning up resources",
            self.__class__.__name__,
            lambda: content.name if isinstance(content := getattr(self, "content", None), Path) else content,
        )

        self.playback.stop()
        self.playback.state.wait_for_cleanup(0, stall_cb=lambda: self.statusLoadingStarted.emit("Clearing buffer..."))
        self.loop.wait_for_threads()

        return super().deleteLater()

    def clear_environment(self) -> None:
        if self._env and not self._env.disposed:
            with self._env.use():
                for cb in self.cbs_on_destroy:
                    self.loop.from_thread(cb)

        self.outputs_manager.clear()

        return super().clear_environment()

    @contextmanager
    def status_loading(self, loading_message: str, completed_message: str) -> Iterator[None]:
        self.statusLoadingStarted.emit(loading_message)
        yield
        self.statusLoadingFinished.emit(completed_message)

    def get_output_metadata(self) -> dict[int, Any]:
        """
        Get metadata for VapourSynth outputs.

        Returns:
            A dictionary mapping output index to metadata string.
        """
        return {}

    @abstractmethod
    def loader(self) -> None: ...

    def init_load(self, frame: int | None = None, tab_index: int | None = None) -> None:
        PluginManager.wait_for_loaded()

    @run_in_background(name="LoadContent")
    def load_content(self, content: T, /, frame: int | None = None, tab_index: int | None = None) -> None:
        logger.debug("load_content called: path=%r, frame=%r, tab_index=%r", content, frame, tab_index)

        self.set_loading_page()
        self.statusLoadingStarted.emit("Loading...")

        self.content = content

        with loader_lock:
            unset_environment()
            self.init_load(frame, tab_index)

            if not (outputs := self._get_outputs()):
                self.clear_failed_load()
                return

            voutputs, aoutputs = outputs

        self.outputs_manager.current_video_index = clamp(self.outputs_manager.current_video_index, 0, len(voutputs) - 1)
        tabs = self.tab_manager.create_tabs(voutputs)

        with QSignalBlocker(self.tab_manager):
            self.tab_manager.swap_tabs(tabs, self.outputs_manager.current_video_index)

        self.tbar.playback_container.set_audio_outputs(aoutputs, self.outputs_manager.current_audio_index)

        # Load plugins in the load_content function so the plugins can get the file_path
        # and do VS things in the init since the environment is already created.
        self.load_plugins()

        @run_in_loop(return_future=False)
        def on_complete(f: Future[None]) -> None:
            if f.exception():
                return

            self.content_area.setEnabled(True)
            self.tab_manager._on_global_autofit_changed(self.tab_manager.autofit_btn.isChecked())
            self.tab_manager.disable_switch = False
            self.playback.can_reload = True

        if not self._on_tab_changed(self.tab_manager.tabs.currentIndex(), cb_render=on_complete):
            logger.error("Failed to load content: %r", self.content)
            self.clear_failed_load()
            return

        logger.info("Content loaded successfully: %r", self.content)
        self.statusLoadingFinished.emit("Completed")

    @run_in_background(name="ReloadContent")
    def reload_content(self) -> None:
        if not self.playback.can_reload:
            logger.warning("Workspace is busy, cannot reload content")
            return

        logger.debug("Reloading content: %r", self.content)

        self.playback.stop()
        self.playback.can_reload = False
        self.statusLoadingStarted.emit("Reloading Content...")

        with self.tbar.disabled(), self.tab_manager.clear_voutputs_on_fail(), self.freeze_viewport():
            self.loop.from_thread(self.content_area.setDisabled, True).result()
            self.tab_manager.disable_switch = True
            self.playback.state.wait_for_cleanup(
                0.25,
                stall_cb=lambda: self.statusLoadingStarted.emit("Clearing buffer..."),
            )

            # 1. Capture state
            saved_state = self.tab_manager.current_view.state

            for view in self.tab_manager.tabs.views():
                self.loop.from_thread(view.clear_scene).result()

            # 2. Reset Environment
            self.clear_environment()
            gc_collect()

            # 3. Load New Content
            with loader_lock:
                if not (outputs := self._get_outputs()):
                    self.clear_failed_load()
                    return

            voutputs, aoutputs = outputs

            # 4. Reconstruct UI
            tabs = self.tab_manager.create_tabs(voutputs, enabled=False)

            # Apply saved pixmap
            for view, voutput in zip(tabs.views(), voutputs, strict=True):
                saved_state.apply_pixmap(view, (voutput.vs_output.clip.width, voutput.vs_output.clip.height))
                saved_state.apply_frozen_state(view)

            with QSignalBlocker(self.tab_manager):
                self.tab_manager.swap_tabs(tabs, clamp(self.tab_manager.tabs.currentIndex(), 0, tabs.count() - 1))

            self.loop.from_thread(
                self.tab_manager._on_global_autofit_changed,
                self.tab_manager.autofit_btn.isChecked(),
                under_reload=True,
            ).result()

            self.tbar.playback_container.set_audio_outputs(aoutputs, self.outputs_manager.current_audio_index)

            @run_in_loop(return_future=False)
            def on_complete(f: Future[None]) -> None:
                if f.exception():
                    return
                self.content_area.setEnabled(True)
                self.tab_manager.tabs.setEnabled(True)
                self.content_area.setFocus()
                self.playback.can_reload = True
                self.tab_manager.disable_switch = False

            if not self._on_tab_changed(
                self.tab_manager.tabs.currentIndex(),
                seamless=True,
                cb_render=on_complete,
                refresh_plugins=True,
            ):
                logger.error("Failed to reload content: %r", self.content)
                self.clear_failed_load()
                raise RuntimeError("Reload failed") from None

            logger.info("Content reloaded successfully: %r", self.content)

    @contextmanager
    def freeze_viewport(self) -> Iterator[None]:
        @run_in_loop(return_future=False)
        def show_screenshot() -> QLabel:
            viewport = self.tab_manager.current_view.viewport()

            screenshot = QLabel(self)
            screenshot.setPixmap(viewport.grab())  # Grab just the viewport, no frame/scrollbars

            pos = viewport.mapTo(self, viewport.rect().topLeft())
            screenshot.setGeometry(pos.x(), pos.y(), viewport.width(), viewport.height())
            screenshot.raise_()
            screenshot.show()
            return screenshot

        screenshot = show_screenshot()

        try:
            yield
        finally:
            self.loop.from_thread(screenshot.setVisible, False)
            self.loop.from_thread(screenshot.deleteLater)
            del screenshot

    @run_in_loop(return_future=False)
    def clear_failed_load(self) -> None:
        self.playback.stop()
        self.playback.state.wait_for_cleanup(0, stall_cb=lambda: self.statusLoadingStarted.emit("Clearing buffer..."))

        with QSignalBlocker(self.tab_manager.tabs):
            self.tab_manager.tabs.clear()

        self.clear_environment()

        self.statusLoadingErrored.emit("Error while loading content")
        self.set_error_page()
        gc_collect()

    @run_in_loop
    def init_timeline(self) -> None:
        if not (voutput := self.outputs_manager.current_voutput):
            logger.debug("No voutput available")
            return

        fps = voutput.vs_output.clip.fps
        total_frames = voutput.vs_output.clip.num_frames

        self.tbar.set_data(total_frames, voutput.cum_durations)

        # Use configured FPS history size, or auto-calculate from FPS when set to 0
        if (fps_history_size := self.global_settings.playback.fps_history_size) <= 0:
            fps_history_size = (
                round(fps)
                if fps > 0
                else round(1 / (sum(voutput.framedurs) / len(voutput.framedurs)))
                if voutput.framedurs
                else 25
            )

        self.playback.state.fps_history = deque(maxlen=clamp(fps_history_size, 1, total_frames))

    @run_in_loop
    def update_timeline_cursor(self, n: int) -> None:
        if not self.outputs_manager.current_voutput:
            return

        self.tbar.timeline.cursor_x = (n := Frame(n))

        with QSignalBlocker(self.tbar.playback_container.frame_edit):
            self.tbar.playback_container.frame_edit.setValue(n)

        with QSignalBlocker(self.tbar.playback_container.time_edit):
            time = self.outputs_manager.current_voutput.frame_to_time(n)
            self.tbar.playback_container.time_edit.setTime(time.to_qtime())

    @run_in_loop(return_future=False)
    def set_loaded_page(self) -> None:
        self.stack.setCurrentWidget(self.loaded_page)
        self.content_area.setFocus()

    @run_in_loop(return_future=False)
    def set_loading_page(self) -> None:
        logger.debug("Switching to loading page")
        self.stack.setCurrentWidget(self.loading_page)

    @run_in_loop(return_future=False)
    def set_empty_page(self) -> None:
        self.stack.setCurrentWidget(self.empty_page)

    @run_in_loop(return_future=False)
    def set_error_page(self) -> None:
        self.stack.setCurrentWidget(self.error_page)

    def _get_outputs(self) -> tuple[list[VideoOutput], list[AudioOutput]] | None:
        try:
            with self.env.use():
                self.loader()

                voutputs = self.outputs_manager.create_voutputs(
                    self.content,
                    self.video_outputs,
                    self.get_output_metadata(),
                    self.api,
                    last_frame=self.playback.state.current_frame,
                )

                if not voutputs:
                    raise RuntimeError

                aoutputs = self.outputs_manager.create_aoutputs(
                    self.content,
                    self.audio_outputs,
                    self.get_output_metadata(),
                    self.api,
                    delay_s=self.tbar.playback_container.audio_delay,
                )

                return voutputs, aoutputs
        except Exception:
            logger.log(DEBUG - 1, "Full traceback:", exc_info=True)
            return None

    def _on_tab_changed(
        self,
        index: int,
        seamless: bool = False,
        cb_render: Callable[[Future[None]], None] | None = None,
        refresh_plugins: bool = False,
    ) -> bool:
        self.playback.stop()
        self.outputs_manager.current_video_index = index

        if not self.outputs_manager.current_voutput:
            logger.debug("Invalid tab index %d, ignoring", index)
            return True

        logger.debug("Switching to video output: clip=%r", self.outputs_manager.current_voutput.vs_output.clip)

        target_frame = self._calculate_target_frame()
        self.init_timeline()
        self.update_timeline_cursor(target_frame)
        self._emit_output_info()

        previous_voutput = self.outputs_manager.voutputs[self.tab_manager.tabs.previous_tab_index]
        current_voutput = self.outputs_manager.current_voutput

        if (
            not (previous_voutput.last_frame == current_voutput.last_frame == target_frame)
            or not current_voutput.loaded_once
        ):
            if not seamless:
                if (prev_pixmap := self.tab_manager.previous_view.pixmap_item.pixmap()).isNull():
                    self.set_loading_page()
                else:
                    self.loaded_page.setDisabled(True)
                    self.tab_manager.update_current_view(prev_pixmap.toImage()).result()
            current_voutput.loaded_once = True

            def on_complete(f: Future[None]) -> None:
                if not f.exception():
                    self.loaded_page.setEnabled(True)
                    self.set_loaded_page()

                    if cb_render:
                        cb_render(f)

                    with self.env.use():
                        self.api._on_current_voutput_changed(refresh_plugins)

            logger.debug("Requesting frame %d", target_frame)
            try:
                self.playback.request_frame(target_frame, on_complete)
            except Exception:
                return False
        else:
            with self.env.use():
                self.api._on_current_voutput_changed(refresh_plugins)

        return True

    def _calculate_target_frame(self, state: PlayHeadToolButton.State | None = None) -> int:
        if not (voutput := self.outputs_manager.current_voutput):
            raise NotImplementedError

        match s := fallback(state, self.tab_manager.sync_playhead_state):
            case PlayHeadToolButton.State.UNLINK:
                target_frame = voutput.last_frame
                logger.debug("Sync playhead %r, using last frame %d", s, target_frame)
                return target_frame
            case PlayHeadToolButton.State.LINK_TIME:
                src_fps = self.outputs_manager.voutputs[self.tab_manager.tabs.previous_tab_index]
                tgt_fps = self.outputs_manager.current_voutput

                current_time = self.outputs_manager.current_voutput.frame_to_time(
                    self.playback.state.current_frame,
                    src_fps,
                )
                target_frame = self.outputs_manager.current_voutput.time_to_frame(
                    current_time,
                    tgt_fps,
                )

                target_frame = clamp(
                    target_frame,
                    0,
                    self.outputs_manager.current_voutput.vs_output.clip.num_frames - 1,
                )

                logger.debug(
                    "Sync playhead %r, targeting frame %d (from time %.3fs)",
                    s,
                    target_frame,
                    current_time.total_seconds(),
                )
                return target_frame
            case PlayHeadToolButton.State.LINK_FRAME:
                target_frame = clamp(
                    self.playback.state.current_frame,
                    0,
                    self.outputs_manager.current_voutput.vs_output.clip.num_frames - 1,
                )
                logger.debug(
                    "Sync playhead %r, targeting frame %d (from frame %d)",
                    s,
                    target_frame,
                    self.playback.state.current_frame,
                )
                return target_frame
            case PlayHeadToolButton.State.LINK_ADAPT as s:
                logger.debug("Sync playhead %r", s)
                resolved_state = {
                    "frame": PlayHeadToolButton.State.LINK_FRAME,
                    "time": PlayHeadToolButton.State.LINK_TIME,
                }[self.tbar.timeline.mode]
                return self._calculate_target_frame(resolved_state)
            case _:
                assert_never(s)

    def _emit_output_info(self) -> None:
        if not (voutput := self.outputs_manager.current_voutput):
            logger.warning("No current video output, ignoring")
            return

        self.statusOutputChanged.emit(voutput.info._replace(sar=self.tab_manager.current_view.display_sar))

    def _on_reload_failed(self) -> None:
        self.load_content(self.content)

    def _copy_current_frame_to_clipboard(self) -> None:
        frame = self.tbar.playback_container.frame_edit.value()

        QApplication.clipboard().setText(str(frame))

        self.statusLoadingFinished.emit(f"Copied frame {frame}")
        logger.info("Copied frame %d to clipboard", frame)

    def _copy_current_time_to_clipboard(self) -> None:
        timestamp = self.tbar.playback_container.time_edit.time().toString("H:mm:ss.zzz")

        QApplication.clipboard().setText(timestamp)

        self.statusLoadingFinished.emit(f"Copied time {timestamp}")
        logger.info("Copied time %s to clipboard", timestamp)

    @run_in_loop(return_future=False)
    def load_plugins(self) -> None:
        if not self.plugins_loaded:
            self.plugins.clear()
            self.docks.clear()

            with self.env.use():
                self._setup_docks()
                self._setup_panels()

            self.plugins_loaded = True

    def _setup_docks(self) -> None:
        for plugin_type in PluginManager.tooldocks:
            dock = PluginDock(plugin_type.display_name, plugin_type.identifier, self.dock_container)
            plugin_obj = plugin_type(dock, self.api)

            dock.setWidget(plugin_obj)
            dock.visibilityChanged.connect(lambda visible, d=dock: self._on_dock_visibility_changed(visible, d))

            self.plugins.append(plugin_obj)
            self.docks.append(dock)

            self.dock_container.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, dock)

            if len(self.docks) > 1:
                self.dock_container.tabifyDockWidget(self.docks[0], dock)

        # Docks are hidden by default, so toggle button starts unchecked
        self.dock_toggle_btn.setChecked(False)

    def _setup_panels(self) -> None:
        for i, plugin_type in enumerate(PluginManager.toolpanels):
            plugin_obj = plugin_type(self.plugin_splitter.plugin_tabs, self.api)

            self.plugins.append(plugin_obj)
            self.plugin_splitter.add_plugin(plugin_obj, plugin_type.display_name)
            self.plugin_splitter.plugin_tabs.setTabVisible(
                i, self.global_settings.view_tools.panels.get(plugin_type.identifier, True)
            )

    def _on_dock_toggle(self, checked: bool) -> None:
        for dock in self.docks:
            if self.global_settings.view_tools.docks.get(dock.objectName(), True):
                dock.setVisible(checked)

    def _on_dock_visibility_changed(self, visible: bool, dock: PluginDock) -> None:
        if not isinstance(w := dock.widget(), WidgetPluginBase):
            return

        if visible:
            with self.env.use():
                self.api._init_plugin(w)

            dock.truly_visible = True
        elif dock.truly_visible:
            w.on_hide()
            dock.truly_visible = False

    def _sync_toolpanel_btn(self, visible: bool) -> None:
        with QSignalBlocker(self.tab_manager.toggle_toolpanel_btn):
            self.tab_manager.toggle_toolpanel_btn.setChecked(visible)

    def _on_splitter_visibility_changed(self, visible: bool = True) -> None:
        if not isinstance(w := self.plugin_splitter.plugin_tabs.currentWidget(), WidgetPluginBase):
            return

        if visible:
            if self.outputs_manager.current_voutput:
                with self.env.use():
                    self.api._init_plugin(w)
        else:
            w.on_hide()

    def _on_splitter_tab_changed(self, new_index: int, old_index: int) -> None:
        if isinstance(w := self.plugin_splitter.plugin_tabs.widget(new_index), WidgetPluginBase):
            with self.env.use():
                self.api._init_plugin(w)

        if isinstance(w := self.plugin_splitter.plugin_tabs.widget(old_index), WidgetPluginBase):
            w.on_hide()


class VSEngineWorkspace[T](LoaderWorkspace[T]):
    """Base workspace for script execution."""

    content_type: ClassVar[Literal["script", "code"]]
    """The type of content to load."""

    script: Script[ManagedEnvironment]
    """The loaded script. Available only after loader() is called."""

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.vsargs = dict[str, Any]()

    @property
    def _script_content(self) -> Any:
        """Return the content to be loaded by the script engine."""
        return self.content

    @property
    def _script_kwargs(self) -> dict[str, Any]:
        """Return additional keyword arguments for vsengine.vpy.load_{content_type}()."""
        return {}

    @property
    def _user_script_path(self) -> str:
        """Return the user script path/filename for error reporting."""
        return (
            str(self._script_content)
            if self.content_type == "script"
            else self._script_kwargs.get("filename", repr(self.content))
        )

    def loader(self) -> None:
        module = ModuleType("__vsview__")
        module.__dict__.update(self.vsargs)

        match self.content_type:
            case "script":
                chdir = Path(self._user_script_path).parent if self.global_settings.chdir else None
                self.script = load_script(
                    self._script_content,
                    self.env,
                    module=module,
                    chdir=chdir,
                    **self._script_kwargs,
                )
            case "code":
                self.script = load_code(self._script_content, self.env, module=module, **self._script_kwargs)
            case _:
                assert_never(self.content_type)

        logger.debug("Running Script...")

        fut = self.script.run()

        try:
            fut.result()
            logger.debug("%s execution completed successfully", self.content_type.title())
        except ExecutionError as e:
            from ...app.error import show_error

            self.statusLoadingErrored.emit("Execution error")

            show_error(e, self, self._user_script_path)
            # Clear traceback to release VS core references held in the exception chain
            e.parent_error.__traceback__ = None
            e.__traceback__ = None

            raise RuntimeError("Script execution failed") from None
